from Globals import InitializeClass
from AccessControl import ClassSecurityInfo
from OFS.SimpleItem import SimpleItem # non e' un folder o altri oggetti simili
from zLOG import LOG, DEBUG, ERROR, INFO, PROBLEM
import Acquisition
from DateTime import DateTime

#from Acquisition import aq_base
# CMF imports
from Products.CMFCore.utils import UniqueObject, getToolByName
#from Products.CMFCore.ActionProviderBase import ActionProviderBase # fornisce azioni
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.Expression import Expression
from Products.CMFCore.CMFCorePermissions import ManagePortal

from Products.PageTemplates.PageTemplateFile import PageTemplateFile

# other imports
import urllib2
from urllib import urlencode
import md5
import time

# security premission

perm = "CMFPayPalIPN: Use PayPal IPN Tool"

# class!

class PayPalIPNTool(UniqueObject, SimpleItem, Acquisition.Implicit):
    """Simple Tool for using IPN PayPal System.
    
    The PayPal Instant Payment Notification is used
    in order to activate a server-side script on the

    seller' site in order to update the informations
    about a transaction. For example, the script is
    called to perform a database update.

    1. first, a paypal server do a POST request to the
    seller' site; in this case we call the portal_pp_tool/confirm
    action

    2. the portal_pp_tool try to resend all the received
    variables making a request to https://www.paypal.com/cgi-bin/webscr
    (SSL connection), adding a "cmd" variable with value
    "_notify-validate"

    3. the paypal server reply with a message, including a
    VERIFIED/NOT VALID Response as message's body, to
    portal_pp_tool

    4. portal_pp_tool must check the some variables, 
    such as payment_status ("Eseguito", "In sospeso",
    "Non eseguito"), txn_id (must be unique),
    receiver_email; usally, we can choose to call a
    validation script (requires cmfformcontroller),
    then a controller script to end the process.

    5. In addiction, portal_pp_tool use a register to
    trace all the pending transactions. The controller
    and validation scripts can check the existence of
    pending transactions and add/remove them.

    for many infos, see https://www.paypal.com/it/cgi-bin/webscr?cmd=p/xcl/rec/ipn-techview-outside
    
    """

    ### Properties ###
    id = "portal_pp_tool"
    meta_type = "CMF PayPal IPN Tool"
    #__implements__ = (ActionProviderBase.__implements__,)
    paypal_url = "https://www.sandbox.paypal.com/cgi-bin/webscr"

    control_script = ""
    action_script = ""
    pending_codes = {}
    expiration_delta = 86400 #pending transaction's minimum expiration time (24h)
    checktime = 3600
    delta = 1
    gamma = 1
    latestcheck = None

    # ZMI tabs
    security = ClassSecurityInfo()
    manage_options = ({ 'label': 'Scripts', 'action': 'manage_scripts'}, ) + \
                     SimpleItem.manage_options

    security.declareProtected( ManagePortal, 'manage_scripts' )
    manage_scripts = PageTemplateFile("zmi/configureScripts", globals(), __name__="manage_scripts") 
    

    ### Actions ###

    ### Methods ###

    security.declareProtected(perm, "confirm")
    def confirm(self, REQUEST, **kwargs):
        """
        Collect the PayPal POST requests and
        resend to paypal_url all the infos
    	"""

        LOG("CMFPayPalIPNTool", INFO, "Try to confirm a payment request...")

        if not REQUEST:
            LOG("CMFPayPalIPNTool", INFO, "No REQUEST found")
            return

        #REQUEST.RESPONSE.write(str(REQUEST.form))

        # new Request object
        params = REQUEST.form
        params ['cmd'] = "_notify-validate"
        req = urllib2.Request(self.paypal_url, urlencode(params))
        req.add_header("Content-type", "application/x-www-form-urlencoded")
        req.add_header("Accept", "text/plain")
        LOG('CMFPayPalIPNTool', DEBUG, "%s" % params)

        # open the connection
        #conn = urllib2.urlopen(self.paypal_url)
        conn = urllib2.urlopen(req)

        status = conn.read()
        LOG('CMFPayPalIPNTool', INFO, "%s" % status)
        if status == "VERIFIED":
            LOG('CMFPayPalIPNTool', INFO, 'Successful payment')
            portal_url = getToolByName(self, 'portal_url')
            portal_obj = portal_url.getPortalObject()
            cntl = portal_obj[self.control_script]
            #cntl = getattr(self.aq_parent, self.control_script)
            actn = portal_obj[self.action_script] 
            if cntl:
                cntl_result = cntl()
                LOG("CMFPayPalIPNTool", INFO, "Invoked control_script: %s" %cntl_result)
                if cntl_result and actn:
                    actn_result = actn()
                    LOG("CMFPayPalIPNTool", INFO, "Invoked action_script: %s" %actn_result)

        elif status == "INVALID":
            LOG('CMFPayPalIPNTool', INFO, 'Not valid payment. Pay Attention!')
        return status

    security.declareProtected(ManagePortal, "getScripts")
    def getScripts(self):
        """Return control_script and action_script values (string)"""
        return {"action_script": self.action_script, "control_script": self.control_script}

    security.declareProtected(ManagePortal, "getUrl")
    def getUrl(self):
        """Return URL value (string)"""
        return self.paypal_url

    security.declareProtected(ManagePortal, "setScripts")
    def setScripts(self, REQUEST):
        """Set script name values"""
        form = REQUEST.form
        self.action_script = form['action_script']
        self.control_script = form['control_script']

        REQUEST.RESPONSE.redirect("%s/manage_scripts" % self.absolute_url())

    security.declareProtected(perm, "setUrl")
    def setUrl(self, REQUEST):
        """Set paypal URL value"""
        form = REQUEST.form
        self.paypal_url = form['paypal_url']

        REQUEST.RESPONSE.redirect("%s/manage_scripts" % self.absolute_url())

    security.declareProtected(perm, "setPending")
    def setPending(self, user, datetime):
        """Register a pending transaction and generate
           a new md5 for it. Returns the md5 digest"""
        LOG("CMFPayPalIPNTool", DEBUG, "Register a new pending code")
        self.autoClean()
        if not isinstance(datetime, DateTime):
            dt = DateTime(datetime)
        else:
            dt = datetime
        LOG("CMFPayPalIPNTool", DEBUG, "Registration time: %s" % dt.timeTime())
        md = md5.new()
        md.update(user)
        md.update(str(datetime.Date()))
        if not self.pending_codes.has_key(user):
            self.pending_codes.setdefault(user, [])
        self.pending_codes[user].append({'datetime': dt,
                                         'md5': md })
        return md.hexdigest()


    security.declareProtected(perm, "getPending")
    def getPending(self, md5, user=None):
        """Check if exists a pending transaction"""
        LOG("CMFPayPalIPNTool", DEBUG, "getPending!")
        if user and self.pending_codes.has_key(user):
            for data in self.pending_codes[user]:
                if data['md5'].hexdigest() == md5:
                    return True
        elif md5:
            for usersdata in self.pending_codes.values():
                for data in usersdata:
                    if data['md5'].hexdigest() == md5:
                        return True
        #else:
        return False

    security.declareProtected(perm, "confirmPending")
    def confirmPending(self, md5, user=None):
        """Pop the pending transaction out of the register"""
        LOG("CMFPayPalIPNTool", DEBUG, "Popping out element from pending_codes")
        if user and self.pending_codes.has_key(user):
            for data in self.pending_codes[user]:
                if data['md5'].hexdigest() == md5:
                    return self.pending_codes[user].remove(data)
        elif md5:
            for user, usersdata in self.pending_codes.items():
                for data in usersdata:
                    if data['md5'].hexdigest() == md5:
                        return user, usersdata.remove(data)
        #else:
        return None

    security.declareProtected(perm, "flushPending")
    def flushPending(self):
        """Clear the pending transactions' register"""
        LOG("CMFPayPalIPNTool", DEBUG, "Clear register")
        self.pending_codes={}
        return True
            
    security.declareProtected(perm, 'index_html')
    index_html = confirm

    def autoClean(self):
        if not self.latestcheck:
            self.latestcheck = time.time() 
        actual = time.time()
        if self.checktime <= 0:
            LOG("CMFPayPalIPNTool", DEBUG, "autoClean started")
            debug = 0
            for usersdata in self.pending_codes.values():
                for data in usersdata:
                    registered = data['datetime']
                    if actual - registered.timeTime() >= self.expiration_delta:
                        usersdata.remove(data)
                        debug = debug + 1
            self.checktime = 3600
            self.latestcheck = actual
            self.gamma = 1
            LOG("CMFPayPalIPNTool", DEBUG, "Cleaned %s items" % debug)
            gap = 0
        else: #calculate the next check interval
            gap = int(self.latestcheck - actual) * 24
            self.checktime = int( self.checktime + gap * self.delta / (self.gamma * self.checktime + 1 ) )
            #self.delta = (self.delta + 1) % 3  + 1
            self.delta = pow(self.gamma, 2)
            self.gamma = self.gamma + 1 
        LOG("CMFPayPalIPNTool", DEBUG, "latestcheck: %s, actual: %s, gap: %s" % (self.latestcheck, actual, gap))
        LOG("CMFPayPalIPNTool", DEBUG, "checktime: %s, delta: %s" % (self.checktime, self.delta))



InitializeClass(PayPalIPNTool)
